#
#  A service meant to handle the details of generating the
#  collection & additional metadata JSON contents at ingest
#
#  The plan is to create a small set of REST endpoints which
#  will generate the relevant data, removing that functionality
#  from the workflow system.
#
#  Additional Metadata: These are files filled with metadata
#       which are not readily extractable from the product to
#       to be ingested, but must be copied over from a predecessor
#
#       NOTE: Currently this is only happens for image ingestion,
#               but that will likely expand over time.
#    Current Uses:
#       VLASS Quicklook Data
#       AOI Ingestion Upon QA Pass
#
#    Future Cases:
#       VLASS cumulative images
#       VLASS cubes
#
#   Collection Metadata: There are set 'collections' within the
#       archive with their own set of specialized metadata.  These
#       fields are specified (and hopefully provided) by the
#       team generating the collection products.
#
#       Since we have been given the VLASS software to maintain, this
#       this tool will be how we interface with their software & database
#       to collect the necessary data for persistence.
#
from pyramid.view import view_config
from pyramid.httpexceptions import exception_response
from pyat.schema import ExecutionBlock, ScienceProduct
from pyat.schema.vlassmodel import ProductVersionArchiveid, ProductVersion, Product, Prerequisite
from sqlalchemy.orm import aliased
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from services import DB
#
# Fundamentally, the additional metadata we need for images are things we already know
# about the EBs used in the creation of that image.
#
#   project_code
#   band_code
#   configurations
#
# Several other fields are simply handled with defaults:
#
#   starttime
#   endtime
#   exposure_time
#   rms_noise
#   image_tags
#   product_tags  (probably more appropriate to use 'set_tags'?)
#
#   calibration_level=2
#
# There is one which needs to be considered in how it works in the new scheme:
#
#   collection_name
#
#


@view_config(route_name='image_metadata_service', renderer='json')
def image_metadata_service(request):
    """Outside contact point for the additional metadata service:
            provided either a MOUS uid or a VLASS Product.vVersion
            string, perform the necessary queries to extract the
            metadata not readily available at image ingestion time."""

    #precursor = ('uid://evla/execblock/593d5051-75d0-4720-8163-5af40e6533e0',)

    input_value = request.params.get('identifier')
    print(f"Received: {input_value}")

    if input_value is not None:
        precursors = find_execblock_locators(input_value)
    else:
        return exception_response(400, details=f"No identifier provided.  Cannot proceed.")

    # Errors signals from our lookup:
    # Yes, I'm having a bit too much fun.
    if -1 == precursors:
        # no results found
        return exception_response(416, details=f"No query results found for {input_value}")
    elif -100 == precursors:
        return exception_response(416, details=f"Multiple query results found for {input_value}")

    # We've got good locators, do the last step and send the JSON:
    return create_additional_metadata_for(precursors)


def find_rawdata_for_quicklook(set_name):
    """We're given the quicklook's cached directory name (basically a product name and version mashed up) and
    must perform the query to get back to the execution block's external name (via the VLASS DB) and then
    obtain the SPL for that EB"""

    # A bit of preprocessing:
    # Quicklook directories combine the name & version of the image.  We need only the early
    # part so perform a split (undisruptive to a product-name-only call).
    set_components = set_name.split('.v')
    product_name=set_components[0]

    # SELECT archive_id FROM product_version_archiveids
    # JOIN product_version ON product_version_archiveids.product_version_id = product_version.id
    # JOIN prerequisite eb ON product_version.product_id = eb.req_product_id
    # JOIN prerequisite cal ON eb.product_id = cal.req_product_id
    # JOIN product ql ON cal.product_id = ql.id
    # WHERE product_version_archiveids.name = 'SdmId'
    # AND ql.name='VLASS1.1.ql.T28t03.J053208+713000.10.2048';

    print(f"Looking backward from Quicklook Product {product_name}")

    eb  = aliased(Prerequisite)
    cal = aliased(Prerequisite)
    ql  = aliased(Product)

    try:
        # We expect only a single EB for QL & SE Images
        answer = DB['VLASS'].query(ProductVersionArchiveid.archive_id)\
            .join(ProductVersion)\
            .join(eb,  ProductVersion.product_id == eb.req_product_id)\
            .join(cal, eb.product_id == cal.req_product_id)\
            .join(ql,  cal.product_id == ql.id)\
            .filter(ProductVersionArchiveid.name == 'SdmId')\
            .filter(ql.name == product_name)\
            .one()

        # Single External name: Single SPL
        locator = DB['SDM'].query(ScienceProduct.science_product_locator)\
            .filter(ScienceProduct.external_name == answer[0])\
            .filter(ScienceProduct.external_system == 'VLA Operations')\
            .one()

    except MultipleResultsFound:
        print(f"Found multiple potential EBs for {product_name}")
        answer = 'Too Many!'
        locator = -100
    except NoResultFound:
        print(f"No eb data found for {product_name}")
        answer = 'No Results!'
        locator = -1

    print(f"For {product_name} the initial EB is {answer} ({locator})")

    return locator


def find_mous_constituents(set_name):
    """Query wrapper for the SPLs of the EBs that make up an MOUS."""
    locator_set = DB['SDM'].query(ExecutionBlock.science_product_locator)\
        .filter(ExecutionBlock.alma_ous_id==set_name)\
        .all()

    if len(locator_set) == 1 and locator_set[0][0] is None:
        # a -1 return value indicates no results (for later return status)
        return -1
    else:
        # unpack the list-of-singleton-tuples into just a list:
        locator_list = unpack_query_results(locator_set)

    print(f"For {set_name} we found {locator_list}")

    # Final result should be a tuple (not list) for interaction
    # with the later queries:
    return tuple(locator_list)


def find_execblock_locators(set_name):
    """Check the input we're given, and determine the
    appropriate product locators to use to build the
    metadata.
    :return a tuple of spls for execblocks"""

    #
    # We expect as input either a VLASS directory name:
    #   VLASS1.2_T19t12.J093318+333000_P48351v1_2019_08_21T16_35_13.601
    # Or an ALMA MOUS UID:
    #   uid://A001/X133d/X1e64
    #
    if 'uid://' in set_name:
        # AOI lookup results in tuple of SPLs
        locators = find_mous_constituents(set_name)
    else:
        # VLASS results in a single SPL already wrapped in a tuple:
        locators = find_rawdata_for_quicklook(set_name)

    # Return the locator(s) in Tuple form to play nice with in_() calls:
    return locators


def create_default_metadata():
    """Fill in the dictionary slots to make sure everything is there.  """
    default_dictionary = {}

    # Defaults on necessary pieces:
    default_dictionary['project_code'] = None
    default_dictionary['band_code'] = None
    default_dictionary['configurations'] = None

    # Static Fields:

    # time fields
    default_dictionary['starttime'] = None
    default_dictionary['endtime']   = None
    default_dictionary['exposure_time'] = None
    default_dictionary['rms_noise'] = None

    # tag & collection fields
    default_dictionary['image_tags'] = ''
    default_dictionary['product_tags'] = ''
    default_dictionary['collection_name'] = ''

    # calibration_level
    default_dictionary['calibration_level'] = 2

    return default_dictionary


def create_additional_metadata_for(locator_list):
    """Provided a list of EB project locators, this aggregates the metadata into
    for all the products into strings for persistence. """
    metadata = create_default_metadata()
    comma = ','

    # Series of 3 queries, with some string manipulation of the results.
    project_tmp = DB['SDM'].query(ExecutionBlock.project_code) \
        .filter(ExecutionBlock.science_product_locator.in_(locator_list)) \
        .distinct() \
        .all()
    project_list = unpack_query_results(project_tmp)
    metadata['project_code'] = comma.join(project_list)

    band_tmp = DB['SDM'].query(ExecutionBlock.band_code) \
        .filter(ExecutionBlock.science_product_locator.in_(locator_list)) \
        .distinct() \
        .all()
    band_list = unpack_query_results(band_tmp)
    metadata['band_code'] = comma.join(band_list)

    cnfg_tmp = DB['SDM'].query(ExecutionBlock.configuration) \
        .filter(ExecutionBlock.science_product_locator.in_(locator_list)) \
        .distinct() \
        .all()

    cnfg_list = unpack_query_results(cnfg_tmp)

    if len(cnfg_list) == 1 and cnfg_list[0] is None:
        # It's possible that ALMA data doesn't have any configuration information:
        print(f"No configuration information available for {locator_list}, blanking the field")
        metadata['configurations'] = ' '
    else:
        metadata['configurations'] = comma.join(cnfg_list)

    return metadata


def unpack_query_results(list_of_tuples):
    """Utility to strip the (,) around the results from an SQL Alchemy Query of a single column"""
    return [value for (value,) in list_of_tuples]

#
# In a similar vein: Look up & return the locator of the calibration used.
# This allows for the creation of the product group structure.
#
# AOI: simple external name query
# VLASS: external name via vlass DB & prereqs, then external name query
#


@view_config(route_name='image_precursor_service', renderer='json')
def image_precursor_service(request):
    """
    Obtain the product locator of the calibration used in the
    imaging run to be ingested (required for appropriate
    provenance tracking via product groups)

    :param request: Web request object
    :return: The desired product locator (or an error)
    """

    input_value = request.params.get('identifier')
    print(f"Received: {input_value}")

    if input_value is not None:
        cal_locator = find_calibration_locator(input_value)
    else:
        return exception_response(400, detail=f"No identifier provided.")

    # If something went wrong with the queries, we get None
    if cal_locator is None:
        return exception_response(416, detail=f"No Calibration locator found for {input_value}")

    # Package up the result:
    return {'locator': cal_locator[0]}


def find_quicklook_calibration_name(set_name):
    """
    We need only look back one set of prerequistes, (see the raw data
    query above) but need to be careful of multiple versions (but
    there should only be one with an ACCEPTED (manual or auto) state).

    Equivalent Query:
    SELECT archive_id FROM product_version_archiveids
    JOIN product_version pv on product_version_archiveids.product_version_id = pv.id
    JOIN prerequisite p on pv.product_id = p.req_product_id
    JOIN product p2 on p.product_id = p2.id
    WHERE p2.name='VLASS1.1.ql.T28t03.J053208+713000.10.2048'
    AND pv.status IN (4,7);

    Answer Format:
        {"VLASS1.1_T28t03.T28t04_P14548v1_2017_10_15T11_40_12.135.tar":"134528"}

    We just want that tarfile name, so we'll have to parse it out.

    :param set_name:  Product name of interest
    :return:          Name of the prereq (disentagled from the JSON)
    """
    #
    # VLASS set names are the product names version information added in:
    #
    product_name = set_name.split('.v')[0]


    try:
        cal_json = DB['VLASS'].query(ProductVersionArchiveid.archive_id)\
                    .join(ProductVersion)\
                    .join(Prerequisite, ProductVersion.product_id==Prerequisite.req_product_id)\
                    .join(Product, Prerequisite.product_id==Product.id)\
                    .filter(ProductVersion.status.in_([4,7]))\
                    .filter(Product.name==product_name)\
                    .one()
    except MultipleResultsFound:
        print(f"Found multiple accepted calibrations for {set_name} ({product_name})")
        return None
    except NoResultFound:
        print(f"No accepted calibration found for {set_name} ({product_name})")
        return None

    #
    # Unwrap the answer:
    #
    json_token = cal_json[0]

    #
    # Now, the json token looks like: {"NameWeNeed":"number"}
    #
    modified_tar_name = json_token.split(':')[0]
    tarfile = modified_tar_name.replace('{', '').replace('"', '')
    print(f"Calibration name: {tarfile}")

    return tarfile


def get_locator_for_external_name(name):
    """
    Simple translator from external name to locator for
    use in creating the ingestion manifest for images.



    :param name: External ID value
    :return:     Project Locator for that ID
    """

    #
    # Workaround:  Some of the UIDs are sanitized in the external names, some not.
    #
    sanitized_name = name.replace(':', '_').replace('/', '_')

    try:
        # torn about this query (not screening by the system)
        locator = DB['SDM'].query(ScienceProduct.science_product_locator)\
                .filter(ScienceProduct.external_name.like(sanitized_name))\
                .order_by(ScienceProduct.metadata_ingestion_date.desc())\
                .first()
    except NoResultFound:
        print(f"No locator found for {name}")
        return None

    return locator


def find_calibration_locator(set_name):
    """For the provided image set identifier, sort out which method to
    use in order to acquire the product locator of the calibration used
    in that imaging run """

    # vlass requires extra work first
    if 'uid://' not in set_name:
        # vlass, go look up the prereq
        external_name = find_quicklook_calibration_name(set_name)
        # if something went wrong, notify the level above
        if external_name is None:
            return None
    else:
        # For AOI, we're using the MOUS UID, and that is used as the calibration's
        # external name for official calibrations.  No further work required.
        external_name = set_name

    # external name -> locator lookup
    return get_locator_for_external_name(external_name)
